"""Main class."""

import opentaskpy.otflogging
from opentaskpy.config.loader import ConfigLoader
from opentaskpy.config.schemas import (
    validate_batch_json,
    validate_execution_json,
    validate_transfer_json,
)
from opentaskpy.taskhandlers.batch import Batch
from opentaskpy.taskhandlers.execution import Execution
from opentaskpy.taskhandlers.transfer import Transfer

GLOBAL_VERBOSITY = 1


class TaskRun:  # pylint: disable=too-few-public-methods
    """Do the actual work.

    Class responsible for doing all the work. The TaskRun class
    parses config, loads variables and triggers the work
    """

    def __init__(self, task_id: str, config_dir: str, noop: bool = False) -> None:
        """Create the TaskRun object.

        Initialises the logging.

        Args:
            task_id (str): ID of the task being triggered
            config_dir (str): Path to the config directory
            noop (bool, optional): Whether to actually run the task or not. If set to
                True, will only check the config loads OK and then exit. Defaults to
                False.
        """
        self.logger = opentaskpy.otflogging.init_logging(__name__, task_id)
        self.task_id = task_id
        self.config_dir = config_dir
        self.active_task_definition = None
        self.noop = noop
        # Create a config loader object
        self.config_loader = ConfigLoader(self.config_dir)

    def run(self) -> bool:
        """Run the task.

        Load all variables, validate the task definitions, and
        trigger the run of the task itself.

        Returns:
            bool: The result of running the task, True is good, False is bad
        """
        global_variables = self.config_loader.get_global_variables()

        # Populate the task definition with the global variables
        active_task_definition = self.config_loader.load_task_definition(self.task_id)

        # Handle any overrides from the command line

        result = False

        # Now we've loaded the config, determine what to do with it
        if "type" not in active_task_definition:
            self.logger.error("Invalid task configuration. Cannot continue")
            return False

        if active_task_definition["type"] == "transfer":
            # Hand off to the transfer module
            self.logger.info("Starting task of type Transfer")
            # Validate the schema
            if not validate_transfer_json(active_task_definition):
                self.logger.error("JSON format does not match schema")
                return False

            if self.noop:
                self.logger.info("Noop set, exiting")
                return True

            transfer = Transfer(global_variables, self.task_id, active_task_definition)

            try:
                result = transfer.run()
            except Exception as exception:  # pylint: disable=broad-except
                transfer.return_result(1, str(exception), exception)

        elif active_task_definition["type"] == "execution":
            # Hand off to the execuiton module
            self.logger.info("Starting task of type Execution")

            # Validate the schema
            if not validate_execution_json(active_task_definition):
                self.logger.error("JSON format does not match schema")
                return False

            if self.noop:
                self.logger.info("Noop set, exiting")
                return True

            execution = Execution(
                global_variables, self.task_id, active_task_definition
            )
            try:
                result = execution.run()
            except Exception as exception:  # pylint: disable=broad-except
                execution.return_result(1, str(exception), exception)

        elif active_task_definition["type"] == "batch":
            # Hand off to the batch module
            self.logger.info("Starting task of type Batch")

            # Validate the schema
            if not validate_batch_json(active_task_definition):
                self.logger.error("JSON format does not match schema")
                return False

            if self.noop:
                self.logger.info("Noop set, exiting")
                return True

            batch = Batch(
                global_variables,
                self.task_id,
                active_task_definition,
                self.config_loader,
            )

            try:
                result = batch.run()
            except Exception as exception:  # pylint: disable=broad-except
                batch.return_result(1, str(exception), exception)

        else:
            self.logger.error("Unknown task type!")

        self.logger.info(f"Task completed with result: {result}")
        # Ensure that the log is closed
        opentaskpy.otflogging.close_log_file(self.logger, result)
        return result



# pylint: skip-file
# ruff: noqa
import pytest
from pytest_shell import fs

from opentaskpy.plugins.lookup.file import run


def test_file_plugin_missing_path():
    with pytest.raises(Exception) as ex:
        run()

    assert (
        ex.value.args[0]
        == "Missing kwarg: 'path' while trying to run lookup plugin 'file'"
    )


def test_file_plugin_file_not_found(tmpdir):
    with pytest.raises(FileNotFoundError) as ex:
        run(path=f"{tmpdir}/does_not_exist.txt")

    assert (
        ex.value.args[0]
        == f"File {tmpdir}/does_not_exist.txt does not exist while trying to run lookup"
        " plugin 'file'"
    )


def test_file_plugin(tmpdir):
    # Run test with a valid variable file, and ensure it's read and contains that value
    file_name = f"{tmpdir}/test.variable.txt"
    content = "test1234"
    fs.create_files([{file_name: {"content": "test1234"}}])
    result = run(path=file_name)
    assert result == content



"""Utility functions for caching variables."""

from importlib import import_module

DEFAULT_CACHE_PLUGINS = ["file"]


def update_cache(cacheable_variable: dict, updated_value: str) -> None:
    """Update the cache with the new value.

    Args:
        cacheable_variable (dict): The cacheable variable to update.
        updated_value (str): The new value to update the cache with.

    Returns:
        None
    """
    # Now find and call the appropriate caching plugin
    handler_package = None
    if cacheable_variable["cachingPlugin"] in DEFAULT_CACHE_PLUGINS:
        handler_package = (
            f"opentaskpy.config.variablecaching.{cacheable_variable['cachingPlugin']}"
        )

    else:
        handler_package = (
            f"opentaskpy.variablecaching.{cacheable_variable['cachingPlugin']}"
        )

    # Call the run method of the handler_package, getting the actual function first
    handler_function = getattr(import_module(handler_package), "run")

    # Call the function
    kwargs = cacheable_variable["cacheArgs"]
    kwargs["value"] = updated_value
    handler_function(**kwargs)



# pylint: skip-file
import pytest

from opentaskpy.config.schemas import validate_transfer_json


@pytest.fixture(scope="function")
def valid_protocol_definition():
    return {"name": "email"}


@pytest.fixture(scope="function")
def valid_source_definition():
    return {
        "hostname": "{{ HOST_A }}",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "test"}},
    }


@pytest.fixture(scope="function")
def valid_destination_definition(valid_protocol_definition):
    return {
        "recipients": ["test@example.com"],
        "subject": "Here is your email",
        "protocol": valid_protocol_definition,
    }


def test_email_basic(valid_source_definition, valid_destination_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
        "destination": [valid_destination_definition],
    }

    assert validate_transfer_json(json_data)



"""A bunch of exceptions that can be raised by the opentaskpy package."""

# mypy: ignore-errors


class FileTooNewError(Exception):
    """File too new error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class LogWatchTimeoutError(Exception):
    """Log watch timeout error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class LogWatchInitError(Exception):
    """Log watch init error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class RemoteTransferError(Exception):
    """Remote transfer error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class RemoteFileNotFoundError(Exception):
    """Remote file not found error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class FilesDoNotMeetConditionsError(Exception):
    """Files do not meet conditions error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class DuplicateConfigFileError(Exception):
    """Duplicate config file error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class InvalidConfigError(Exception):
    """Invalid config error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class UnknownProtocolError(Exception):
    """Unknown protocol error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class LookupPluginError(Exception):
    """Lookup plugin error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class CachingPluginError(Exception):
    """Caching plugin error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class VariableResolutionTooDeepError(Exception):
    """Variable resolution too deep error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class SSHClientError(Exception):
    """SSH client error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class DecryptionNotSupportedError(Exception):
    """Decryption not supported error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class EncryptionNotSupportedError(Exception):
    """Encryption not supported error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class EncryptionError(Exception):
    """Generic encryption error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)


class DecryptionError(Exception):
    """Generic decryption error."""

    def __init__(self, message):
        """Call the base class constructor."""
        super().__init__(message)



# pylint: skip-file
# ruff: noqa

import pytest

from opentaskpy.remotehandlers.ssh import SSHTransfer


def test_cacheable_variable_dotted_notation():

    spec = {"task_id": "1234", "x": {"y": "value"}}
    rh = SSHTransfer(spec)

    assert rh.obtain_variable_from_spec("x.y", spec) == "value"


def test_cacheable_variable_dotted_notation_array():

    spec = {
        "task_id": "1234",
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*taskhandler.proxy\\.txt",
        "invalidParam": [1, 2, 3, 4, 5],
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    }
    rh = SSHTransfer(spec)

    assert rh.obtain_variable_from_spec("invalidParam[2]", spec) == "3"



# pylint: skip-file
# ruff: noqa
import hashlib
import os
import random
from copy import deepcopy

import gnupg
import pytest
from pytest_shell import fs

from opentaskpy import exceptions
from opentaskpy.taskhandlers import transfer
from tests.fixtures.pgp import *  # noqa: F403, F401
from tests.fixtures.ssh_clients import *  # noqa: F403, F401

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"


local_test_dir = "/tmp/local_tests"

# Create a task definition
local_task_definition = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "local"},
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
            "mode": "0644",
        },
    ],
}

local_file_watch_task_no_error_definition = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": ".*nofileexists.*\\.txt",
        "fileWatch": {"timeout": 1},
        "error": False,
        "protocol": {"name": "local"},
    },
}

local_with_fin_task_definition = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": ".*taskhandler.fin.*\\.txt",
        "protocol": {"name": "local"},
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "flags": {
                "fullPath": f"{local_test_dir}/dest/sftp_with_fin.fin",
            },
            "protocol": {"name": "local"},
        },
    ],
}

# PCA delete
local_pca_delete_task_definition_1 = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "pca_delete\\.txt",
        "protocol": {"name": "local"},
        "postCopyAction": {
            "action": "delete",
        },
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

# PCA move
local_pca_move_task_definition_1 = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "pca_move\\.txt",
        "protocol": {"name": "local"},
        "postCopyAction": {
            "action": "move",
            "destination": f"{local_test_dir}/archive",
        },
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

local_pca_move_task_definition_2 = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "pca_move_2\\.txt",
        "protocol": {"name": "local"},
        "postCopyAction": {
            "action": "move",
            "destination": f"{local_test_dir}/archive/",
        },
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

local_pca_invalid_move_task_definition = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "pca_move_3\\.txt",
        "protocol": {"name": "local"},
        "postCopyAction": {
            "action": "move",
            "destination": f"{local_test_dir}/archive/pca_move_bad.txt",
        },
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

local_pca_invalid_move_dir_task_definition = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "pca_move_4\\.txt",
        "protocol": {"name": "local"},
        "postCopyAction": {
            "action": "move",
            "destination": "/etc/passwd",
        },
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

local_pca_rename_task_definition_1 = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "pca_rename_1\\.txt",
        "protocol": {"name": "local"},
        "postCopyAction": {
            "action": "rename",
            "destination": f"{local_test_dir}/archive/",
            "pattern": "rename",
            "sub": "renamed",
        },
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

local_pca_rename_many_task_definition_1 = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "pca_rename_many.*\\.txt",
        "protocol": {"name": "local"},
        "postCopyAction": {
            "action": "rename",
            "destination": f"{local_test_dir}/archive/",
            "pattern": "rename",
            "sub": "renamed",
        },
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}


local_destination_file_rename = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "dest_rename.*taskhandler.*\\.txt",
        "protocol": {"name": "local"},
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "rename": {
                "pattern": "t(askha)ndler",
                "sub": "T\\1NDLER",
            },
            "protocol": {"name": "local"},
        },
    ],
}

fail_invalid_protocol_task_definition = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "nonexistent"},
    },
}

local_multi_protocol_task_definition = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "local"},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
        {
            "hostname": "172.16.0.22",
            "directory": "/home/application/testFiles/dest",
            "protocol": {"name": "sftp", "credentials": {"username": "application"}},
        },
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

# Count conditional tests
local_task_with_counts = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "counts[0-9]\\.txt",
        "conditionals": {
            "count": {
                "minCount": 2,
                "maxCount": 2,
            },
        },
        "protocol": {"name": "local"},
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}

local_file_watch_task_with_counts = {
    "type": "transfer",
    "source": {
        "directory": f"{local_test_dir}/src",
        "fileRegex": "counts_watch[0-9]\\.txt",
        "fileWatch": {"timeout": 5},
        "conditionals": {
            "count": {
                "minCount": 2,
                "maxCount": 2,
            },
            "checkDuringFilewatch": True,
        },
        "protocol": {"name": "local"},
    },
    "destination": [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ],
}


@pytest.fixture(scope="session")
def setup_local_test_dir():
    os.makedirs(f"{local_test_dir}/src", exist_ok=True)
    os.makedirs(f"{local_test_dir}/dest", exist_ok=True)
    os.makedirs(f"{local_test_dir}/archive", exist_ok=True)

    return local_test_dir


def test_remote_handler():
    # Validate that given a transfer with local protocol, that we get a remote handler of type local

    transfer_obj = transfer.Transfer(None, "sftp-basic", local_task_definition)

    transfer_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert transfer_obj.source_remote_handler.__class__.__name__ == "LocalTransfer"

    # dest_remote_handler should be an array
    assert isinstance(transfer_obj.dest_remote_handlers, list)
    assert len(transfer_obj.dest_remote_handlers) == 1
    #  of LocalTransfer objects
    assert transfer_obj.dest_remote_handlers[0].__class__.__name__ == "LocalTransfer"


def test_local_non_existent_file(setup_local_test_dir):
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["fileRegex"] = ".*nonexistent.*\\.txt"
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-non-existent", local_task_definition_copy
    )

    # Run the transfer and expect a FilesDoNotMeetConditionsError exception
    with pytest.raises(exceptions.FilesDoNotMeetConditionsError):
        transfer_obj.run()


def test_local_basic(setup_local_test_dir):
    # Create a test file
    fs.create_files(
        [{f"{local_test_dir}/src/test.taskhandler.txt": {"content": "test1234"}}]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "local-basic", local_task_definition)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.taskhandler.txt")

    # Generate a random number
    random_number = random.randint(1, 1000000)
    # Test transferring to a subdir that doesn't exist, and creating it
    local_task_definition["destination"][0][
        "directory"
    ] = f"/{local_test_dir}/dest/{random_number}"

    transfer_obj = transfer.Transfer(None, "local-basic", local_task_definition)

    # Run the transfer and expect a false status, as we've not asked the directory to be created
    # Expect a RemoteTransferError
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()

    assert not os.path.exists(
        f"{local_test_dir}/dest/{random_number}/test.taskhandler.txt"
    )

    # Check file mode is 644
    assert (
        oct(os.stat(f"{local_test_dir}/dest/test.taskhandler.txt").st_mode)[-3:]
        == "644"
    )

    # Now run again, but ask for the dir to be created

    local_task_definition["destination"][0]["createDirectoryIfNotExists"] = True

    transfer_obj = transfer.Transfer(None, "local-basic", local_task_definition)

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/{random_number}/test.taskhandler.txt")


def test_local_filewatch_no_error():
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-no-file-no-error", local_file_watch_task_no_error_definition
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()


def test_local_basic_write_fin(setup_local_test_dir):
    # Delete any fin files that exist
    for file in os.listdir(f"{local_test_dir}/dest"):
        if file.endswith(".fin"):
            os.remove(f"{local_test_dir}/dest/{file}")

    # Create a test file
    fs.create_files(
        [{f"{local_test_dir}/src/test.taskhandler.fin.txt": {"content": "test1234"}}]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-basic", local_with_fin_task_definition
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.taskhandler.fin.txt")

    # Check the fin file exists
    assert os.path.exists(f"{local_test_dir}/dest/sftp_with_fin.fin")


def test_pca_move(setup_local_test_dir):
    # Empty the PCA archive directory
    for file in os.listdir(f"{local_test_dir}/archive"):
        os.remove(f"{local_test_dir}/archive/{file}")

    # Create the test file
    fs.create_files([{f"{local_test_dir}/src/pca_move.txt": {"content": "test1234"}}])
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-pca-move", local_pca_move_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/pca_move.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{local_test_dir}/src/pca_move.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{local_test_dir}/archive/pca_move.txt")

    # Create the next test file
    fs.create_files([{f"{local_test_dir}/src/pca_move_2.txt": {"content": "test1234"}}])
    transfer_obj = transfer.Transfer(
        None, "local-pca-move-2", local_pca_move_task_definition_2
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/pca_move_2.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{local_test_dir}/src/pca_move_2.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{local_test_dir}/archive/pca_move_2.txt")

    fs.create_files([{f"{local_test_dir}/src/pca_move_3.txt": {"content": "test1234"}}])
    transfer_obj = transfer.Transfer(
        None, "local-pca-move-invalid", local_pca_invalid_move_task_definition
    )

    # Run the transfer and expect a failure status
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/pca_move_3.txt")
    # Check the source file should still exist
    assert os.path.exists(f"{local_test_dir}/src/pca_move_3.txt")

    # Check the source file has not been archived
    assert not os.path.exists(f"{local_test_dir}/archive/pca_move_bad.txt")

    # Finally, try moving to a directory that doesn't exist
    fs.create_files([{f"{local_test_dir}/src/pca_move_4.txt": {"content": "test1234"}}])
    transfer_obj = transfer.Transfer(
        None, "local-pca-move-invalid", local_pca_invalid_move_dir_task_definition
    )
    # Run the transfer and expect a failure status
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/pca_move_4.txt")
    # Check the source file should still exist
    assert os.path.exists(f"{local_test_dir}/src/pca_move_4.txt")

    # Check the source file has not been archived
    assert not os.path.exists(
        f"{local_test_dir}/archive/archive_no-exist/pca_move_bad.txt"
    )


def test_pca_delete(setup_local_test_dir):
    # Create the test file
    fs.create_files([{f"{local_test_dir}/src/pca_delete.txt": {"content": "test1234"}}])
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-pca-delete", local_pca_delete_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/pca_delete.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{local_test_dir}/src/pca_delete.txt")


def test_destination_file_rename(setup_local_test_dir):
    # Random number
    import random

    random_no = random.randint(1, 1000)

    # Create the test file
    fs.create_files(
        [
            {
                f"{local_test_dir}/src/dest_rename_{random_no}_taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-dest-rename", local_destination_file_rename
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(
        f"{local_test_dir}/dest/dest_rename_{random_no}_TaskhaNDLER.txt"
    )


def test_pca_rename(setup_local_test_dir):
    # Empty the PCA archive directory
    for file in os.listdir(f"{local_test_dir}/archive"):
        os.remove(f"{local_test_dir}/archive/{file}")

    # Create the test file
    fs.create_files(
        [{f"{local_test_dir}/src/pca_rename_1.txt": {"content": "test1234"}}]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-pca-rename", local_pca_rename_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{local_test_dir}/dest/pca_rename_1.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{local_test_dir}/src/pca_rename_1.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{local_test_dir}/archive/pca_renamed_1.txt")


def test_pca_rename_many(setup_local_test_dir):
    # Create the test file
    # Empty the PCA archive directory
    for file in os.listdir(f"{local_test_dir}/archive"):
        os.remove(f"{local_test_dir}/archive/{file}")

    # for 1 to 10
    for i in range(1, 10):
        fs.create_files(
            [{f"{local_test_dir}/src/pca_rename_many_{i}.txt": {"content": "test1234"}}]
        )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-pca-rename-name", local_pca_rename_many_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    for i in range(1, 10):
        assert os.path.exists(f"{local_test_dir}/dest/pca_rename_many_{i}.txt")
        # Check the source file no longer exists
        assert not os.path.exists(f"{local_test_dir}/src/pca_rename_many_{i}.txt")

        # Check the source file has been archived
        assert os.path.exists(f"{local_test_dir}/archive/pca_renamed_many_{i}.txt")


def test_local_multi_protocol(
    root_dir, setup_local_test_dir, setup_ssh_keys, setup_sftp_keys
):
    # Create a test file
    fs.create_files(
        [{f"{local_test_dir}/src/test.taskhandler.txt": {"content": "test1234"}}]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-multi-protocol", local_multi_protocol_task_definition
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test.taskhandler.txt")
    assert os.path.exists(f"{root_dir}/testFiles/sftp_2/dest/test.taskhandler.txt")
    assert os.path.exists(f"{local_test_dir}/dest/test.taskhandler.txt")


def test_local_decrypt_incoming_file(
    tmpdir, root_dir, setup_local_test_dir, private_key, public_key
):

    # Create a test file
    fs.create_files(
        [{f"{tmpdir}/src/test.decryption.txt": {"content": "test12345678"}}]
    )

    # Checksum the file
    with open(f"{tmpdir}/src/test.decryption.txt", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(public_key)

    # Encrypt the file
    with open(f"{tmpdir}/src/test.decryption.txt", "rb") as f:
        status = gpg.encrypt_file(
            f,
            always_trust=True,
            recipients="test@example.com",
            output=f"{tmpdir}/test.decryption.txt.gpg",
        )

        # Check status returns a 0 exit code
        assert status.ok

    # Check the output file exists
    assert os.path.exists(f"{tmpdir}/test.decryption.txt.gpg")

    # Now we have an encrypted file (we can pretend we are collecting from elsewhere),
    # run a transfer to copy the file locally, and decrypt it
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{tmpdir}"
    local_task_definition_copy["source"]["fileRegex"] = "test.decryption.txt.gpg"
    local_task_definition_copy["source"]["encryption"] = {
        "decrypt": True,
        "private_key": private_key,
    }

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "local-decrypt", local_task_definition_copy)
    assert transfer_obj.run()

    # Check the decrypted source files have been deleted
    assert not os.path.exists(f"{tmpdir}/test.decryption.txt")

    # Check the output file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.decryption.txt")
    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{local_test_dir}/dest/test.decryption.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum


def test_local_decrypt_incoming_file_custom_extensions(
    tmpdir, root_dir, setup_local_test_dir, private_key, public_key
):

    # Create a test file
    fs.create_files(
        [{f"{tmpdir}/src/test.decryption.txt": {"content": "test12345678"}}]
    )

    # Checksum the file
    with open(f"{tmpdir}/src/test.decryption.txt", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(public_key)

    # Encrypt the file
    with open(f"{tmpdir}/src/test.decryption.txt", "rb") as f:
        status = gpg.encrypt_file(
            f,
            always_trust=True,
            recipients="test@example.com",
            output=f"{tmpdir}/test.decryption.txt.pgp",
        )

        # Check status returns a 0 exit code
        assert status.ok

    # Check the output file exists
    assert os.path.exists(f"{tmpdir}/test.decryption.txt.pgp")

    # Now we have an encrypted file (we can pretend we are collecting from elsewhere),
    # run a transfer to copy the file locally, and decrypt it
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{tmpdir}"
    local_task_definition_copy["source"]["fileRegex"] = "test.decryption.txt.pgp"
    local_task_definition_copy["source"]["encryption"] = {
        "decrypt": True,
        "private_key": private_key,
    }

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "local-decrypt", local_task_definition_copy)
    assert transfer_obj.run()

    # Check the decrypted source files have been deleted
    assert not os.path.exists(f"{tmpdir}/test.decryption.txt")

    # Check the output file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.decryption.txt")
    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{local_test_dir}/dest/test.decryption.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum

    # Now do it again, but with a different extension that's not pgp or gpg
    # Rename the original encrypted file and use that as the input
    os.rename(f"{tmpdir}/test.decryption.txt.pgp", f"{tmpdir}/test.decryption.txt.enc")
    local_task_definition_copy["source"]["fileRegex"] = "test.decryption.txt.enc"

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "local-decrypt", local_task_definition_copy)
    assert transfer_obj.run()

    # Check the decrypted source files have been deleted
    assert not os.path.exists(f"{tmpdir}/test.decryption.txt")

    # Check the output file exists with the .decrypted file extension
    assert os.path.exists(f"{local_test_dir}/dest/test.decryption.txt.enc.decrypted")


def test_local_encrypt_outgoing_file(
    tmpdir, root_dir, setup_local_test_dir, private_key, public_key
):

    # Create a test file
    fs.create_files(
        [{f"{tmpdir}/src/test.encryption.txt": {"content": "test12345678"}}]
    )

    # Checksum the file
    with open(f"{tmpdir}/src/test.encryption.txt", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(private_key)

    # run a transfer to copy the file locally, and encrypt it
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{tmpdir}/src"
    local_task_definition_copy["source"]["fileRegex"] = "test.encryption.txt"

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
            "encryption": {
                "encrypt": True,
                "public_key": public_key,
            },
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "local-encrypt", local_task_definition_copy)
    assert transfer_obj.run()

    # Check the output file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.encryption.txt.gpg")

    # Now we need to decrypt this file to check that it matches the original content,
    # and can actually be decrypted again
    decryption_data = gpg.decrypt_file(
        open(f"{local_test_dir}/dest/test.encryption.txt.gpg", "rb"),
        output=f"{local_test_dir}/dest/test.encryption.txt",
    )

    assert decryption_data.ok
    assert decryption_data.returncode == 0

    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{local_test_dir}/dest/test.encryption.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum

    # Ensure that the source encrypted file has been deleted
    assert not os.path.exists(f"{local_test_dir}/src/test.encryption.txt.gpg")


def test_local_encrypt_outgoing_file_custom_extension(
    tmpdir, root_dir, setup_local_test_dir, private_key, public_key
):

    # Create a test file
    fs.create_files(
        [{f"{tmpdir}/src/test.encryption_custom_ext.txt": {"content": "test12345678"}}]
    )

    # Checksum the file
    with open(f"{tmpdir}/src/test.encryption_custom_ext.txt", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(private_key)

    # run a transfer to copy the file locally, and encrypt it
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{tmpdir}/src"
    local_task_definition_copy["source"]["fileRegex"] = "test.encryption_custom_ext.txt"

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
            "encryption": {
                "encrypt": True,
                "public_key": public_key,
                "output_extension": "pgp",
            },
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "local-encrypt", local_task_definition_copy)
    assert transfer_obj.run()

    # Check the output file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.encryption_custom_ext.txt.pgp")

    # Now we need to decrypt this file to check that it matches the original content,
    # and can actually be decrypted again
    decryption_data = gpg.decrypt_file(
        open(f"{local_test_dir}/dest/test.encryption_custom_ext.txt.pgp", "rb"),
        output=f"{local_test_dir}/dest/test.encryption_custom_ext.txt",
    )

    assert decryption_data.ok
    assert decryption_data.returncode == 0

    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{local_test_dir}/dest/test.encryption_custom_ext.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum

    # Ensure that the source encrypted file has been deleted
    assert not os.path.exists(
        f"{local_test_dir}/src/test.encryption_custom_ext.txt.pgp"
    )


def test_local_encrypt_with_signing_outgoing_file(
    tmpdir, root_dir, setup_local_test_dir, private_key, public_key
):

    # Create a test file
    fs.create_files(
        [{f"{tmpdir}/src/test.encryptionSign.txt": {"content": "test12345678"}}]
    )

    # Checksum the file
    with open(f"{tmpdir}/src/test.encryptionSign.txt", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the private key
    import_result = gpg.import_keys(private_key)

    # run a transfer to copy the file locally, and encrypt it
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{tmpdir}/src"
    local_task_definition_copy["source"]["fileRegex"] = "test.encryptionSign.txt"

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
            "encryption": {
                "encrypt": True,
                "sign": True,
                "private_key": private_key,
                "public_key": public_key,
            },
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(
        None, "local-encrypt-and-sign", local_task_definition_copy
    )
    assert transfer_obj.run()

    # Check the output file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.encryptionSign.txt.gpg")

    # Now we need to decrypt this file to check that it matches the original content,
    # and can actually be decrypted again
    decryption_data = gpg.decrypt_file(
        open(f"{local_test_dir}/dest/test.encryptionSign.txt.gpg", "rb"),
        output=f"{local_test_dir}/dest/test.encryptionSign.txt",
    )

    assert decryption_data.ok
    assert decryption_data.returncode == 0

    # Check that it was signed to start with
    assert decryption_data.sig_info is not None
    # Get the sig_info, loop through the signatures and check that they all have a status of "signature valid"
    for sig in decryption_data.sig_info:
        assert decryption_data.sig_info[sig]["status"] == "signature valid"

    # Check that the file's checksum matches that of the original unencrypted source file
    # Check the checksum of the new file
    with open(f"{local_test_dir}/dest/test.encryptionSign.txt", "rb") as f:
        new_file_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_file_checksum == original_file_checksum

    # Ensure that the source encrypted file has been deleted
    assert not os.path.exists(f"{local_test_dir}/src/test.encryptionSign.txt.gpg")


def test_local_encrypt_with_signing_missing_key_outgoing_file(
    tmpdir,
    root_dir,
    setup_local_test_dir,
    private_key,
    public_key_2,
    private_key_2,
    public_key,
):

    # Create a test file
    fs.create_files(
        [{f"{tmpdir}/src/test.encryptionSign2.txt": {"content": "test12345678"}}]
    )

    # Checksum the file
    with open(f"{tmpdir}/src/test.encryptionSign2.txt", "rb") as f:
        original_file_checksum = hashlib.md5(f.read()).hexdigest()

    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the second private key
    import_result = gpg.import_keys(private_key_2)

    # run a transfer to copy the file locally, and encrypt it
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{tmpdir}/src"
    local_task_definition_copy["source"]["fileRegex"] = "test.encryptionSign2.txt"

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
            "encryption": {
                "encrypt": True,
                "sign": True,
                "private_key": private_key,
                "public_key": public_key_2,
            },
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(
        None, "local-encrypt-and-sign", local_task_definition_copy
    )
    assert transfer_obj.run()

    # Check the output file exists
    assert os.path.exists(f"{local_test_dir}/dest/test.encryptionSign2.txt.gpg")

    # Now we need to decrypt this file to check that it matches the original content,
    # and can actually be decrypted again
    decryption_data = gpg.decrypt_file(
        open(f"{local_test_dir}/dest/test.encryptionSign2.txt.gpg", "rb"),
        output=f"{local_test_dir}/dest/test.encryptionSign2.txt",
    )

    # Check that there was a problem and the stratus is signature error
    problems = decryption_data.problems
    # There should be an object in the array with a status of signature error
    assert any("signature error" in problem["status"] for problem in problems)

    assert (
        decryption_data.ok
    )  # Decrypt was ok, but the return code will be non-zero due to sig error
    assert decryption_data.returncode != 0


def test_transfer_decryption_failure_local(
    tmpdir, root_dir, setup_local_test_dir, private_key_2, public_key
):

    # Create a file and encrypt it with public key
    fs.create_files(
        [{f"{tmpdir}/src/test.decryption.txt": {"content": "test12345678"}}]
    )
    # Create a gpg object
    gpg = gnupg.GPG(gnupghome=f"{tmpdir}")

    # Import the public key
    gpg.import_keys(public_key)

    # Encrypt the file
    with open(f"{tmpdir}/src/test.decryption.txt", "rb") as f:
        status = gpg.encrypt_file(
            f,
            always_trust=True,
            recipients="test@example.com",
            output=f"{tmpdir}/test.decryption.txt.gpg",
        )

        # Check status returns a 0 exit code
        assert status.ok

    # Check the output file exists
    assert os.path.exists(f"{tmpdir}/test.decryption.txt.gpg")

    # Now we have an encrypted file (we can pretend we are collecting from elsewhere),
    # run a transfer to copy the file locally, and decrypt it
    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{tmpdir}"
    local_task_definition_copy["source"]["fileRegex"] = "test.decryption.txt.gpg"
    local_task_definition_copy["source"]["encryption"] = {
        "decrypt": True,
        "private_key": private_key_2,
    }

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
        },
    ]

    # Run the transfer
    transfer_obj = transfer.Transfer(None, "local-decrypt", local_task_definition_copy)
    # This should fail to decrypt because the private key is incorrect
    with pytest.raises(exceptions.DecryptionError):
        transfer_obj.run()


def test_transfer_encryption_local_invalid_key(root_dir, setup_local_test_dir):
    # Create a test file
    fs.create_files(
        [{f"{local_test_dir}/src/test.encryption.txt": {"content": "test12345678"}}]
    )

    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{local_test_dir}/src"
    local_task_definition_copy["source"]["fileRegex"] = "test.encryption.txt"

    # Override the destination
    local_task_definition_copy["destination"] = [
        {
            "directory": f"{local_test_dir}/dest",
            "protocol": {"name": "local"},
            "encryption": {
                "encrypt": True,
                "public_key": "invalid_public_key",
            },
        },
    ]

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "local-encrypt", local_task_definition_copy)

    # Run the transfer and expect an exception due to failed encryption
    with pytest.raises(exceptions.EncryptionError):
        transfer_obj.run()


def test_transfer_decryption_local_invalid_key(root_dir, setup_local_test_dir):
    # Create a test file
    fs.create_files(
        [{f"{local_test_dir}/src/test.encryption.txt": {"content": "test12345678"}}]
    )

    local_task_definition_copy = deepcopy(local_task_definition)
    local_task_definition_copy["source"]["directory"] = f"{local_test_dir}/src"
    local_task_definition_copy["source"]["fileRegex"] = "test.encryption.txt"
    local_task_definition_copy["source"]["encryption"] = {
        "decrypt": True,
        "private_key": "invalid_private_key",
    }

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "local-decrypt", local_task_definition_copy)

    # Run the transfer and expect an exception due to failed encryption
    with pytest.raises(exceptions.EncryptionError):
        transfer_obj.run()


def test_local_counts():
    # Create a test file
    fs.create_files(
        [
            {f"{local_test_dir}/src/counts1.txt": {"content": "test1234"}},
            {f"{local_test_dir}/src/counts2.txt": {"content": "test1234"}},
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "local-counts", local_task_with_counts)

    # Run the transfer and expect a true status
    assert transfer_obj.run()


def test_local_counts_error():
    # Create a test file
    fs.create_files(
        [
            {f"{local_test_dir}/src/counts_error1.txt": {"content": "test1234"}},
        ]
    )
    local_task_with_counts_error = deepcopy(local_task_with_counts)
    local_task_with_counts_error["source"]["fileRegex"] = "counts_error[0-9]\\.txt"

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None,
        "local-filewatch-counts-error-min",
        local_task_with_counts_error,
    )
    # Test 1 file < minCount of 2 errors
    # Run the transfer and expect a FilesDoNotMeetConditionsError exception
    with pytest.raises(exceptions.FilesDoNotMeetConditionsError):
        transfer_obj.run()

    fs.create_files(
        [
            {f"{local_test_dir}/src/counts_error2.txt": {"content": "test1234"}},
            {f"{local_test_dir}/src/counts_error3.txt": {"content": "test1234"}},
        ]
    )

    transfer_obj = transfer.Transfer(
        None,
        "local-filewatch-counts-error-max",
        local_task_with_counts_error,
    )
    #  Test 3 files > maxCount of 2 errors
    # Run the transfer and expect a FilesDoNotMeetConditionsError exception
    with pytest.raises(exceptions.FilesDoNotMeetConditionsError):
        transfer_obj.run()


def test_local_filewatch_counts():
    # Create a test file
    fs.create_files(
        [
            {f"{local_test_dir}/src/counts_watch1.txt": {"content": "test1234"}},
            {f"{local_test_dir}/src/counts_watch2.txt": {"content": "test1234"}},
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "local-filewatch-counts", local_file_watch_task_with_counts
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()


def test_local_filewatch_counts_error():
    # Create a test file
    fs.create_files(
        [
            {f"{local_test_dir}/src/counts_watch_error1.txt": {"content": "test1234"}},
        ]
    )
    local_file_watch_task_with_counts_error = deepcopy(
        local_file_watch_task_with_counts
    )
    local_file_watch_task_with_counts_error["source"][
        "fileRegex"
    ] = "counts_watch_error[0-9]\\.txt"

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None,
        "local-filewatch-counts-error-min",
        local_file_watch_task_with_counts_error,
    )
    # Test 1 file < minCount of 2 errors
    # Run the transfer and expect a RemoteFileNotFoundError exception
    with pytest.raises(exceptions.RemoteFileNotFoundError):
        transfer_obj.run()

    fs.create_files(
        [
            {f"{local_test_dir}/src/counts_watch_error2.txt": {"content": "test1234"}},
            {f"{local_test_dir}/src/counts_watch_error3.txt": {"content": "test1234"}},
        ]
    )

    transfer_obj = transfer.Transfer(
        None,
        "local-filewatch-counts-error-max",
        local_file_watch_task_with_counts_error,
    )
    #  Test 3 files > maxCount of 2 errors
    # Run the transfer and expect a RemoteFileNotFoundError exception
    with pytest.raises(exceptions.RemoteFileNotFoundError):
        transfer_obj.run()



# pylint: skip-file
# ruff: noqa
import os

import pytest

from opentaskpy import exceptions
from opentaskpy.taskhandlers import execution
from tests.fixtures.ssh_clients import *  # noqa: F403

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

local_test_dir = "/tmp/local_tests"

# Create a task definition
touch_task_definition = {
    "type": "execution",
    "directory": "/tmp",
    "command": f"touch {local_test_dir}/dest/execution.txt",
    "protocol": {"name": "local"},
}

fail_task_definition = {
    "type": "execution",
    "directory": "/tmp",
    "command": f"test -e {local_test_dir}/src/execution.test.fail.txt",
    "protocol": {"name": "local"},
}

fail_host_task_definition = {
    "type": "execution",
    "directory": "/tmp",
    "command": f"touch {local_test_dir}/dest/execution.invalidhost.txt",
    "protocol": {"name": "local"},
}

fail_invalid_protocol_task_definition = {
    "type": "execution",
    "directory": "/tmp",
    "command": f"touch {local_test_dir}/dest/execution.invalidhost.txt",
    "protocol": {"name": "rubbish"},
}


@pytest.fixture(scope="session")
def setup_local_test_dir():
    os.makedirs(f"{local_test_dir}/src", exist_ok=True)
    os.makedirs(f"{local_test_dir}/dest", exist_ok=True)
    os.makedirs(f"{local_test_dir}/archive", exist_ok=True)

    return local_test_dir


def test_invalid_protocol():
    execution_obj = execution.Execution(
        None, "invalid-protocol", fail_invalid_protocol_task_definition
    )
    # Expect a UnknownProtocolError exception
    with pytest.raises(exceptions.UnknownProtocolError):
        execution_obj._set_remote_handlers()


def test_basic_execution(setup_local_test_dir):
    execution_obj = execution.Execution(None, "df-basic", touch_task_definition)
    execution_obj._set_remote_handlers()

    # Ensure no test files exist already, if so delete them
    if os.path.exists(f"{local_test_dir}/dest/execution.txt"):
        os.remove(f"{local_test_dir}/dest/execution.txt")

    # Validate some things were set as expected
    assert execution_obj.remote_handlers[0].__class__.__name__ == "LocalExecution"

    # Run the execution and expect a true status
    assert execution_obj.run()

    # Check the destination file exists on both hosts
    assert os.path.exists(f"{local_test_dir}/dest/execution.txt")


def test_basic_execution_cmd_failure(setup_local_test_dir):
    execution_obj = execution.Execution(None, "task-fail", fail_task_definition)
    execution_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert execution_obj.remote_handlers[0].__class__.__name__ == "LocalExecution"

    # Run the execution and expect a failure

    assert not execution_obj.run()



# pylint: skip-file
# ruff: noqa
import os
from copy import deepcopy
from datetime import datetime

import pytest
from pytest_shell import fs

from opentaskpy import exceptions
from opentaskpy.taskhandlers import transfer
from tests.fixtures.ssh_clients import *  # noqa: F403, F401

os.environ["OTF_NO_LOG"] = "1"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

# Create a task definition
scp_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}


scp_task_definition_no_permissions = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/etc",
            "rename": {
                "pattern": ".*taskhandler.*\\.txt",
                "sub": "passwd",
            },
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

scp_file_watch_task_no_error_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*nofileexists.*\\.txt",
        "fileWatch": {"timeout": 1},
        "error": False,
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
}

scp_with_fin_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*taskhandler.fin.*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "flags": {
                "fullPath": "/tmp/testFiles/dest/scp_with_fin.fin",
            },
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

# PCA move
scp_pca_move_task_definition_1 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "pca_move\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "move",
            "destination": "/tmp/testFiles/archive",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

scp_pca_move_task_definition_2 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "pca_move_2\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "move",
            "destination": "/tmp/testFiles/archive/",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

scp_pca_invalid_move_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "pca_move_3\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "move",
            "destination": "/tmp/testFiles/archive/pca_move_bad.txt",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

scp_pca_rename_task_definition_1 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "pca_rename_1\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "rename",
            "destination": "/tmp/testFiles/archive/",
            "pattern": "rename",
            "sub": "renamed",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

scp_pca_rename_many_task_definition_1 = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "pca_rename_many.*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        "postCopyAction": {
            "action": "rename",
            "destination": "/tmp/testFiles/archive/",
            "pattern": "rename",
            "sub": "renamed",
        },
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

# Proxy task definition
scp_proxy_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*taskhandler.proxy\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "transferType": "proxy",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

scp_destination_file_rename = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "dest_rename.*taskhandler.*\\.txt",
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "rename": {
                "pattern": "t(askha)ndler",
                "sub": "T\\1NDLER",
            },
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

fail_invalid_protocol_task_definition = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": ".*taskhandler.*\\.txt",
        "protocol": {"name": "nonexistent"},
    },
}

# Count conditional tests
scp_task_with_counts = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "counts[0-9]\\.txt",
        "conditionals": {
            "count": {
                "minCount": 2,
                "maxCount": 2,
            },
        },
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}

scp_file_watch_task_with_counts = {
    "type": "transfer",
    "source": {
        "hostname": "172.16.0.11",
        "directory": "/tmp/testFiles/src",
        "fileRegex": "counts_watch[0-9]\\.txt",
        "fileWatch": {"timeout": 5},
        "conditionals": {
            "count": {
                "minCount": 2,
                "maxCount": 2,
            },
            "checkDuringFilewatch": True,
        },
        "protocol": {"name": "ssh", "credentials": {"username": "application"}},
    },
    "destination": [
        {
            "hostname": "172.16.0.12",
            "directory": "/tmp/testFiles/dest",
            "protocol": {"name": "ssh", "credentials": {"username": "application"}},
        },
    ],
}


def test_invalid_protocol():
    transfer_obj = transfer.Transfer(
        None, "invalid-protocol", fail_invalid_protocol_task_definition
    )
    # Expect a UnknownProtocolError exception
    with pytest.raises(exceptions.UnknownProtocolError):
        transfer_obj._set_remote_handlers()

    # Try one with a longer protocol name
    fail_invalid_protocol_task_definition["source"]["protocol"][
        "name"
    ] = "some.module.path.ProtocolClass"
    transfer_obj = transfer.Transfer(
        None, "invalid-protocol", fail_invalid_protocol_task_definition
    )
    # Expect a UnknownProtocolError exception
    with pytest.raises(exceptions.UnknownProtocolError):
        transfer_obj._set_remote_handlers()


def test_remote_handler(setup_ssh_keys):
    # Validate that given a transfer with ssh protocol, that we get a remote handler of type SSH

    transfer_obj = transfer.Transfer(None, "scp-basic", scp_task_definition)

    transfer_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert transfer_obj.source_remote_handler.__class__.__name__ == "SSHTransfer"

    # dest_remote_handler should be an array
    assert isinstance(transfer_obj.dest_remote_handlers, list)
    assert len(transfer_obj.dest_remote_handlers) == 1
    #  of SSHTransfer objects
    assert transfer_obj.dest_remote_handlers[0].__class__.__name__ == "SSHTransfer"


def test_scp_basic(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/test.taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "scp-basic", scp_task_definition)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test.taskhandler.txt")


def test_scp_basic_key_from_protocol_definition(root_dir, ssh_key_file):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/test.taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    scp_task_definition_copy = deepcopy(scp_task_definition)
    scp_task_definition_copy["source"]["protocol"]["credentials"]["key"] = ssh_key_file
    transfer_obj = transfer.Transfer(None, "scp-basic", scp_task_definition_copy)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test.taskhandler.txt")


def test_scp_basic_create_dest_dir(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/test.taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    scp_task_definition_copy = deepcopy(scp_task_definition)
    scp_task_definition_copy["destination"][0][
        "directory"
    ] = f"/tmp/testFiles/{datetime.now().strftime('%s')}"
    scp_task_definition_copy["destination"][0]["createDirectoryIfNotExists"] = True
    transfer_obj = transfer.Transfer(None, "scp-basic", scp_task_definition_copy)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test.taskhandler.txt")


def test_scp_basic_no_permissions(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/test.taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-basic-no-permissions", scp_task_definition_no_permissions
    )

    # Run the transfer and expect a true status
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()


def test_scp_filewatch_no_error(setup_ssh_keys):
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-no-file-no-error", scp_file_watch_task_no_error_definition
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()


def test_scp_basic_write_fin(root_dir, setup_ssh_keys):
    # Delete any fin files that exist
    for file in os.listdir(f"{root_dir}/testFiles/ssh_2/dest"):
        if file.endswith(".fin"):
            os.remove(f"{root_dir}/testFiles/ssh_2/dest/{file}")

    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/test.taskhandler.fin.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "scp-basic", scp_with_fin_task_definition)

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test.taskhandler.fin.txt")

    # Check the fin file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/scp_with_fin.fin")


def test_pca_move(root_dir, setup_ssh_keys):
    # Empty the PCA archive directory
    for file in os.listdir(f"{root_dir}/testFiles/ssh_1/archive"):
        os.remove(f"{root_dir}/testFiles/ssh_1/archive/{file}")

    # Create the test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/pca_move.txt": {"content": "test1234"}}]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-pca-move", scp_pca_move_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/pca_move.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{root_dir}/testFiles/ssh_1/src/pca_move.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/archive/pca_move.txt")

    # Create the next test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/pca_move_2.txt": {"content": "test1234"}}]
    )
    transfer_obj = transfer.Transfer(
        None, "scp-pca-move-2", scp_pca_move_task_definition_2
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()

    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/pca_move_2.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{root_dir}/testFiles/ssh_1/src/pca_move_2.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/archive/pca_move_2.txt")

    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/pca_move_3.txt": {"content": "test1234"}}]
    )
    transfer_obj = transfer.Transfer(
        None, "scp-pca-move-invalid", scp_pca_invalid_move_task_definition
    )

    # Run the transfer and expect a true status
    with pytest.raises(exceptions.RemoteTransferError):
        transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/pca_move_3.txt")
    # Check the source file should still exist
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/src/pca_move_3.txt")

    # Check the source file has not been archived
    assert not os.path.exists(f"{root_dir}/testFiles/ssh_1/archive/pca_move_bad.txt")


def test_destination_file_rename(root_dir, setup_ssh_keys):
    # Random number
    import random

    random_no = random.randint(1, 1000)

    # Create the test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/dest_rename_{random_no}_taskhandler.txt": {
                    "content": "test1234"
                }
            }
        ]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-dest-rename", scp_destination_file_rename
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(
        f"{root_dir}/testFiles/ssh_2/dest/dest_rename_{random_no}_TaskhaNDLER.txt"
    )


def test_pca_rename(root_dir, setup_ssh_keys):
    # Empty the PCA archive directory
    for file in os.listdir(f"{root_dir}/testFiles/ssh_1/archive"):
        os.remove(f"{root_dir}/testFiles/ssh_1/archive/{file}")

    # Create the test file
    fs.create_files(
        [{f"{root_dir}/testFiles/ssh_1/src/pca_rename_1.txt": {"content": "test1234"}}]
    )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-pca-rename", scp_pca_rename_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/pca_rename_1.txt")
    # Check the source file no longer exists
    assert not os.path.exists(f"{root_dir}/testFiles/ssh_1/src/pca_rename_1.txt")

    # Check the source file has been archived
    assert os.path.exists(f"{root_dir}/testFiles/ssh_1/archive/pca_renamed_1.txt")


def test_pca_rename_many(root_dir, setup_ssh_keys):
    # Create the test file
    # Empty the PCA archive directory
    for file in os.listdir(f"{root_dir}/testFiles/ssh_1/archive"):
        os.remove(f"{root_dir}/testFiles/ssh_1/archive/{file}")

    # for 1 to 10
    for i in range(1, 10):
        fs.create_files(
            [
                {
                    f"{root_dir}/testFiles/ssh_1/src/pca_rename_many_{i}.txt": {
                        "content": "test1234"
                    }
                }
            ]
        )
    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "scp-pca-rename-name", scp_pca_rename_many_task_definition_1
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    for i in range(1, 10):
        assert os.path.exists(
            f"{root_dir}/testFiles/ssh_2/dest/pca_rename_many_{i}.txt"
        )
        # Check the source file no longer exists
        assert not os.path.exists(
            f"{root_dir}/testFiles/ssh_1/src/pca_rename_many_{i}.txt"
        )

        # Check the source file has been archived
        assert os.path.exists(
            f"{root_dir}/testFiles/ssh_1/archive/pca_renamed_many_{i}.txt"
        )


def test_scp_proxy(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/test.taskhandler.proxy.txt": {
                    "content": "test1234"
                }
            }
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "scp-basic", scp_proxy_task_definition)
    local_staging_dir = transfer_obj.local_staging_dir

    # Run the transfer and expect a true status
    assert transfer_obj.run()
    # Check the destination file exists
    assert os.path.exists(f"{root_dir}/testFiles/ssh_2/dest/test.taskhandler.proxy.txt")

    # Ensure that local files are tidied up
    assert not os.path.exists(local_staging_dir)


def test_invalid_ssh_decryption_direct():
    # Create a transfer object
    scp_task_definition_copy = deepcopy(scp_task_definition)
    scp_task_definition_copy["source"]["encryption"] = {"decrypt": True}
    transfer_obj = transfer.Transfer(None, "scp-basic", scp_task_definition_copy)

    # Expect a DecryptionNotSupportedError exception
    with pytest.raises(exceptions.DecryptionNotSupportedError):
        transfer_obj.run()


def test_invalid_ssh_encryption_direct():
    # Create a transfer object
    scp_task_definition_copy = deepcopy(scp_task_definition)
    scp_task_definition_copy["destination"][0]["encryption"] = {"encrypt": True}
    transfer_obj = transfer.Transfer(None, "scp-basic", scp_task_definition_copy)

    # Expect a EncryptionNotSupportedError exception
    with pytest.raises(exceptions.EncryptionNotSupportedError):
        transfer_obj.run()


def test_scp_counts(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {f"{root_dir}/testFiles/ssh_1/src/counts1.txt": {"content": "test1234"}},
            {f"{root_dir}/testFiles/ssh_1/src/counts2.txt": {"content": "test1234"}},
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(None, "ssh-counts", scp_task_with_counts)

    # Run the transfer and expect a true status
    assert transfer_obj.run()


def test_scp_counts_error(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_error1.txt": {
                    "content": "test1234"
                }
            },
        ]
    )
    scp_task_with_counts_error = deepcopy(scp_task_with_counts)
    scp_task_with_counts_error["source"]["fileRegex"] = "counts_error[0-9]\\.txt"

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None,
        "ssh-filewatch-counts-error-min",
        scp_task_with_counts_error,
    )
    # Test 1 file < minCount of 2 errors
    # Run the transfer and expect a FilesDoNotMeetConditionsError exception
    with pytest.raises(exceptions.FilesDoNotMeetConditionsError):
        transfer_obj.run()

    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_error2.txt": {
                    "content": "test1234"
                }
            },
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_error3.txt": {
                    "content": "test1234"
                }
            },
        ]
    )

    transfer_obj = transfer.Transfer(
        None,
        "ssh-filewatch-counts-error-max",
        scp_task_with_counts_error,
    )
    #  Test 3 files > maxCount of 2 errors
    # Run the transfer and expect a FilesDoNotMeetConditionsError exception
    with pytest.raises(exceptions.FilesDoNotMeetConditionsError):
        transfer_obj.run()


def test_scp_filewatch_counts(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_watch1.txt": {
                    "content": "test1234"
                }
            },
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_watch2.txt": {
                    "content": "test1234"
                }
            },
        ]
    )

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None, "ssh-filewatch-counts", scp_file_watch_task_with_counts
    )

    # Run the transfer and expect a true status
    assert transfer_obj.run()


def test_scp_filewatch_counts_error(root_dir, setup_ssh_keys):
    # Create a test file
    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_watch_error1.txt": {
                    "content": "test1234"
                }
            },
        ]
    )
    scp_file_watch_task_with_counts_error = deepcopy(scp_file_watch_task_with_counts)
    scp_file_watch_task_with_counts_error["source"][
        "fileRegex"
    ] = "counts_watch_error[0-9]\\.txt"

    # Create a transfer object
    transfer_obj = transfer.Transfer(
        None,
        "ssh-filewatch-counts-error-min",
        scp_file_watch_task_with_counts_error,
    )
    # Test 1 file < minCount of 2 errors
    # Run the transfer and expect a RemoteFileNotFoundError exception
    with pytest.raises(exceptions.RemoteFileNotFoundError):
        transfer_obj.run()

    fs.create_files(
        [
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_watch_error2.txt": {
                    "content": "test1234"
                }
            },
            {
                f"{root_dir}/testFiles/ssh_1/src/counts_watch_error3.txt": {
                    "content": "test1234"
                }
            },
        ]
    )

    transfer_obj = transfer.Transfer(
        None,
        "ssh-filewatch-counts-error-max",
        scp_file_watch_task_with_counts_error,
    )
    #  Test 3 files > maxCount of 2 errors
    # Run the transfer and expect a RemoteFileNotFoundError exception
    with pytest.raises(exceptions.RemoteFileNotFoundError):
        transfer_obj.run()



# pylint: skip-file
import pytest

from opentaskpy.config.schemas import validate_transfer_json


@pytest.fixture(scope="function")
def valid_source_definition():
    return {
        "accessToken": "1234",
        "protocol": {"name": "dummy"},
    }


def test_dummy_basic(valid_source_definition):
    json_data = {"type": "transfer", "source": valid_source_definition}

    assert validate_transfer_json(json_data)


# Test dummy again but with cachableVariables defined too
def test_dummy_with_cachable_variables(valid_source_definition):
    json_data = {
        "type": "transfer",
        "source": valid_source_definition,
    }

    json_data["source"]["cacheableVariables"] = [
        {
            "variableName": "accessToken",
            "cachingPlugin": "file",
            "cacheArgs": {
                "file": "/tmp/cacheable_variable.txt",
            },
        }
    ]

    assert validate_transfer_json(json_data)

    # Remove the cacheArgs and validate the validation fails
    del json_data["source"]["cacheableVariables"][0]["cacheArgs"]
    assert not validate_transfer_json(json_data)



"""SSH Remote Handlers.

This module contains the SSH remote handlers for transfers and executions.
"""

import glob
import logging
import os
import random
import re
import stat
import time
from io import StringIO
from shlex import quote

from paramiko import AutoAddPolicy, RSAKey, SFTPClient, SSHClient, Transport
from paramiko.channel import ChannelFile, ChannelStderrFile
from tenacity import retry, stop_after_attempt, wait_exponential

import opentaskpy.otflogging
from opentaskpy.exceptions import SSHClientError
from opentaskpy.remotehandlers.remotehandler import (
    RemoteExecutionHandler,
    RemoteTransferHandler,
)

SSH_OPTIONS: str = "-o StrictHostKeyChecking=no -o BatchMode=yes -o ConnectTimeout=5"
REMOTE_SCRIPT_BASE_DIR: str = "/tmp"  # nosec B108


class SSHTransfer(RemoteTransferHandler):
    """SSH Transfer Handler."""

    TASK_TYPE = "T"

    FILE_NAME_DELIMITER = "|||"

    ssh_client: SSHClient | None = None
    sftp_connection: SFTPClient | None = None
    log_watch_start_row = 0

    def __init__(self, spec: dict):
        """Initialise the SSHTransfer handler.

        Args:
            spec (dict): The spec for the transfer. This is either the source, or the
            destination spec.
        """
        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        client = SSHClient()
        client.set_log_channel(f"{__name__}.{ spec['task_id']}.paramiko.transport")
        client.set_missing_host_key_policy(AutoAddPolicy())
        self.ssh_client = client

        # Handle default values
        if "createDirectoryIfNotExists" not in spec:
            spec["createDirectoryIfNotExists"] = False

        super().__init__(spec)

    def supports_direct_transfer(self) -> bool:
        """Return True, as SSH allows direct transfers by using the scp command."""
        return True

    def connect(self, hostname: str, ssh_client: SSHClient | None = None) -> None:
        """Connect to the remote host.

        Args:
            hostname (str): The hostname to connect to.
            ssh_client (SSHClient, optional): An existing SSHClient to use. Defaults to None.
        """
        is_remote_host = False
        if ssh_client is not None:
            is_remote_host = True
        else:
            ssh_client = self.ssh_client

        if ssh_client is None:
            self.logger.error(f"[{self.spec['hostname']}] SSH client not initialised")
            raise SSHClientError("SSH Client not initialised")

        if ssh_client.get_transport() and ssh_client.get_transport().is_active():  # type: ignore[union-attr]
            self.logger.debug(
                f"[{self.spec['hostname']}] SSH connection to {hostname} already active"
            )
            return

        kwargs = {
            "hostname": hostname,
            "port": (self.spec["protocol"].get("port", 22)),
            "username": self.spec["protocol"]["credentials"]["username"],
            "timeout": 3,
            "allow_agent": False,
        }
        # If a custom key is set via env vars, then set that
        if (
            os.environ.get("OTF_SSH_KEY")
            and os.path.exists(str(os.environ.get("OTF_SSH_KEY")))
        ) and "keyFile" not in self.spec["protocol"]["credentials"]:
            self.logger.info("Loading custom private SSH key from OTF_SSH_KEY env var")
            key = RSAKey.from_private_key_file(str(os.environ.get("OTF_SSH_KEY")))
            kwargs["pkey"] = key

        # If a specific key file has been defined, then use that
        elif "keyFile" in self.spec["protocol"]["credentials"]:
            self.logger.info("Using key file from task spec")
            kwargs["key_filename"] = self.spec["protocol"]["credentials"]["keyFile"]

        # If a private key has been defined as a string, then use that instead
        elif "key" in self.spec["protocol"]["credentials"]:
            self.logger.info("Using private key from task spec")
            key = RSAKey.from_private_key(
                StringIO(self.spec["protocol"]["credentials"]["key"])
            )
            kwargs["pkey"] = key

        self.connect_with_retry(ssh_client, kwargs)

        _, stdout, _ = ssh_client.exec_command("uname -a")  # nosec B601
        with stdout as stdout_fh:
            self.logger.log(
                11,
                f"[{self.spec['hostname']}] Remote uname:"
                f" {stdout_fh.read().decode('UTF-8')}",
            )

        sftp = ssh_client.open_sftp()

        if not is_remote_host:
            self.sftp_connection = sftp

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
    )
    def connect_with_retry(self, ssh_client: SSHClient, kwargs: dict) -> None:
        """Connect to the remote host with retry.

        Args:
            ssh_client (SSHClient): The SSH client to use.
            kwargs (dict): The keyword arguments to use to connect to the remote host.
        """
        try:
            self.logger.info(f"Connecting to {kwargs['hostname']}")
            ssh_client.connect(**kwargs)

        except Exception as ex:
            self.logger.error(
                f"[{self.spec['hostname']}] Unable to connect to {kwargs['hostname']}: {ex}"
            )
            raise ex

    def tidy(self) -> None:
        """Tidy up the SSH connection.

        Shut down the SSH connection and remove the remote transfer script from the
        remote host.
        """
        # Remove remote scripts
        if self.sftp_connection:

            self.logger.info(f"[{self.spec['hostname']}] Closing SFTP connection")
            self.sftp_connection.get_channel().close()  # type: ignore[union-attr]

            # Wait until the channel is closed
            time.sleep(0.25)
            for _ in range(2):
                if not self.sftp_connection.get_channel().closed:  # type: ignore[union-attr]
                    time.sleep(0.5)
                else:
                    break

            self.sftp_connection.close()

        if self.ssh_client:
            self.logger.info(f"[{self.spec['hostname']}] Closing SSH connection")
            self.ssh_client.close()

    def get_staging_directory(self, remote_spec: dict) -> str:
        """Get the staging directory for the remote host.

        Args:
            remote_spec (dict): The remote spec.

        Returns:
            str: The staging directory on the remote host.
        """
        # Get the user's home directory
        if "stagingDirectory" in remote_spec:
            return str(remote_spec["stagingDirectory"])

        # Check SSH connection is established
        if not self.ssh_client.get_transport().is_active():  # type: ignore[union-attr]
            raise SSHClientError("SSH connection not active")

        _, stdout, _ = self.ssh_client.exec_command("echo $HOME")  # type: ignore[union-attr]  # nosec B601
        with stdout as stdout_fh:
            home_dir = stdout_fh.read().decode("UTF-8").strip()

        return f"{home_dir}/otf/{ self.spec['task_id']}/"

    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Return list of files that match the source definition.

        Args:
            directory (str, optional): The directory to search in. Defaults to None.
            file_pattern (str, optional): The file pattern to search for. Defaults to
            None.

        Returns:
            dict: A dict of files that match the source definition.
        """
        self.connect(self.spec["hostname"])
        if not directory:
            directory = str(self.spec["directory"])
        if not file_pattern:
            file_pattern = str(self.spec["fileRegex"])

        self.logger.info(
            f"Searching for files in {directory} with pattern {file_pattern}"
        )

        self.logger.log(
            12,
            f"[{self.spec['hostname']}] Searching in {directory} for files with"
            f" pattern {file_pattern}",
        )
        remote_files: dict = {}
        # Check the remote directory exists
        try:
            self.sftp_connection.stat(directory)  # type: ignore[union-attr]
        except FileNotFoundError:
            self.logger.error(
                f"[{self.spec['hostname']}] Directory {directory} does not exist"
            )
            return remote_files

        remote_file_list = self.sftp_connection.listdir(directory)  # type: ignore[union-attr]
        for file in list(remote_file_list):
            if re.match(file_pattern, file):
                # Get the file attributes
                file_attr = self.sftp_connection.lstat(f"{directory}/{file}")  # type: ignore[union-attr]
                self.logger.log(12, f"File attributes {file_attr}")
                remote_files[f"{directory}/{file}"] = {
                    "size": file_attr.st_size,
                    "modified_time": file_attr.st_mtime,
                }

        return remote_files

    def pull_files_to_worker(
        self, files: list[str], local_staging_directory: str
    ) -> int:
        """Pull files to the worker.

        This function is used when we need to download source files from the source
        server onto the worker. These are then later pushed to the destination server

        Args:
            files (list): A list of files to download.
            local_staging_directory (str): The local staging directory to download the
            files to.

        Returns:
            int: 0 if successful, 1 if not.
        """
        result = 0
        # Connect to the source
        self.connect(self.spec["hostname"])

        # Create the staging directory locally
        if not os.path.exists(local_staging_directory):
            os.makedirs(local_staging_directory)

        # Download the files via SFTP
        for file in files:
            self.logger.info(
                f"[LOCALHOST] Downloading file {file} to {local_staging_directory}"
            )
            file_name = os.path.basename(file)
            try:
                self.sftp_connection.get(file, f"{local_staging_directory}/{file_name}")  # type: ignore[union-attr]
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(
                    f"[LOCALHOST] Unable to download file locally via SFTP: {ex}"
                )
                result = 1

        return result

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Push files from the worker to the destination server.

        This function is used when the source files have been downloaded locally and
        need to be uploaded to the destination server. This would be expected to be
        called against the remote handler for the destination server.

        Args:
            local_staging_directory (str): The local staging directory to upload the
            files from.
            file_list (dict, optional): A list of files to upload. Defaults to None.

        Returns:
            int: 0 if successful, 1 if not.
        """
        # Connect to the destination server
        self.connect(self.spec["hostname"])
        # Check that the SFTP client is connected and active
        if not isinstance(self.sftp_connection, SFTPClient):
            self.logger.error(f"[{self.spec['hostname']}] Cannot connect via SFTP")
            return 1

        # Handle the staging directory
        destination_directory = self.get_staging_directory(self.spec)

        # Sanitize the destination directory
        destination_directory = quote(destination_directory)

        # Use SFTP connection to check if the directory exists
        try:
            self.sftp_connection.stat(destination_directory)
        except FileNotFoundError:
            # Create the directory
            self.logger.info(
                f"[{self.spec['hostname']}] Creating destination directory"
                f" {destination_directory}"
            )
            mkdir_p(self.sftp_connection, destination_directory)

        # Transfer the files, just use SFTP
        result = 0

        if file_list:
            files = list(file_list.keys())
        else:
            # Get list of files in local_staging_directory
            files = glob.glob(f"{local_staging_directory}/*")
        for file in files:
            self.logger.info(f"[LOCALHOST] Transferring file via SFTP: {file}")
            file_name = os.path.basename(file)
            try:
                self.sftp_connection.put(file, f"{destination_directory}{file_name}")
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(f"[LOCALHOST] Unable to transfer file via SFTP: {ex}")
                result = 1

        return result

    def transfer_files(
        self,
        files: list[str],
        remote_spec: dict,
        dest_remote_handler: RemoteTransferHandler | None = None,
    ) -> int:
        """Transfer files from the source server to the destination server.

        Args:
            files (dict): A dictionary of files to transfer.
            remote_spec (dict): The remote specification for the destination server.
            dest_remote_handler (RemoteTransferHandler, optional): The remote handler
            for the destination server. Defaults to None.

        Returns:
            int: 0 if successful, if not, the return code from the remotely executed SCP
            command.
        """
        remote_host = remote_spec["hostname"]
        self.connect(remote_host)

        # If we are given a destination handler, make sure we connect to the host
        if dest_remote_handler:
            self.connect(remote_host, dest_remote_handler.ssh_client)

        # Construct an SCP command to transfer the files to the destination server
        remote_user = (
            remote_spec["protocol"]["credentials"]["transferUsername"]
            if "transferUsername" in remote_spec["protocol"]["credentials"]
            else remote_spec["protocol"]["credentials"]["username"]
        )

        # Handle staging directory if there is one
        destination_directory = self.get_staging_directory(remote_spec)

        # Check that the SFTP client is connected and active
        dest_sftp_client = dest_remote_handler.ssh_client.open_sftp()

        # Create/validate staging directory exists on destination
        # Use SFTP connection to check if the directory exists
        try:
            dest_sftp_client.stat(destination_directory)
        except FileNotFoundError:
            # Create the directory
            self.logger.info(
                f"[{dest_remote_handler.spec['hostname']}] Creating destination"
                f" directory {destination_directory}"
            )
            mkdir_p(dest_sftp_client, destination_directory)

        # Sanitise arguments
        files = [quote(file) for file in files]
        destination_directory = quote(destination_directory)
        remote_user = quote(remote_user)
        remote_host = quote(remote_host)

        remote_command = f'scp {SSH_OPTIONS} {" ".join(files)} {remote_user}@{remote_host}:"{destination_directory}"'

        self.logger.info(
            f"[{self.spec['hostname']}] Transferring files via SCP: {remote_command}"
        )

        _, stdout, stderr = self.ssh_client.exec_command(remote_command)  # type: ignore[union-attr] # nosec B601

        self._log_remote_output(stdout, stderr)

        remote_rc = stdout.channel.recv_exit_status()
        self.logger.info(
            f"[{self.spec['hostname']}] Got return code {remote_rc} from SCP command"
        )

        return remote_rc

    def pull_files(self, files: list[str], remote_spec: dict) -> int:
        """Pull files from the source server to the destination server.

        Args:
            files (list[str]): A list of files to pull.
            remote_spec (dict): The remote specification for the source server.

        Returns:
            int: 0 if successful, if not, the return code from the remotely executed SCP
        """
        self.connect(self.spec["hostname"])
        # Construct an SCP command to transfer the files from the source server
        source_user = self.spec["protocol"]["credentials"]["transferUsername"]
        source_host = remote_spec["hostname"]

        # Handle staging directory if there is one
        destination_directory = self.get_staging_directory(self.spec)

        # Get an SFTP client to use to create the staging directory
        sftp_client = self.ssh_client.open_sftp()  # type: ignore[union-attr]

        # Create/validate staging directory exists on destination
        # Use SFTP connection to check if the directory exists
        try:
            sftp_client.stat(destination_directory)
        except FileNotFoundError:
            # Create the directory
            self.logger.info(
                f"[{self.spec['hostname']}] Creating destination directory"
                f" {destination_directory}"
            )
            mkdir_p(sftp_client, destination_directory)

        files_str = ""
        for file in files:
            # Sanitise file
            file = quote(file)
            source_host = quote(source_host)
            source_user = quote(source_user)

            files_str += f"{source_user}@{source_host}:{file} "

        destination_directory = quote(destination_directory)

        remote_command = (
            f"scp {SSH_OPTIONS} {files_str.strip()} {destination_directory}"
        )
        self.logger.info(
            f"[{self.spec['hostname']}] Transferring files via SCP: {remote_command}"
        )

        _, stdout, stderr = self.ssh_client.exec_command(remote_command)  # type: ignore[union-attr] # nosec B601

        self._log_remote_output(stdout, stderr)
        remote_rc = stdout.channel.recv_exit_status()
        self.logger.info(
            f"[{self.spec['hostname']}] Got return code {remote_rc} from SCP command"
        )

        return remote_rc

    def move_files_to_final_location(self, files: dict) -> int:
        """Move files from the staging directory to their final location.

        Args:
            files (dict): A dictionary of files to move.

        Returns:
            int: 0 if successful, 1 if not.
        """
        self.connect(self.spec["hostname"])

        # Convert all the source file names into the filename with the destination directory as a prefix
        files_with_directory = []
        for file in list(files):
            files_with_directory.append(
                f"{self.get_staging_directory(self.spec)}{os.path.basename(file)}"
            )
        self.FILE_NAME_DELIMITER.join(files_with_directory).strip()

        directory = self.spec["directory"]

        # Get an SFTP connection if it doesn't exist
        if not isinstance(self.sftp_connection, SFTPClient):
            self.sftp_connection = self.ssh_client.open_sftp()  # type: ignore[union-attr]

        # Check if the destination directory exists on the remote host
        try:
            self.sftp_connection.stat(directory)
        except FileNotFoundError:
            if self.spec["createDirectoryIfNotExists"]:
                # Create the directory
                self.logger.info(
                    f"[{self.spec['hostname']}] Creating destination directory"
                    f" {directory}"
                )
                mkdir_p(self.sftp_connection, directory)
            else:
                self.logger.error(
                    f"[{self.spec['hostname']}] Destination directory {directory}"
                    " does not exist"
                )
                return 1

        # Move the files to the right place and apply any renames and permissions that
        # are needed
        for file in list(files):
            current_path = (
                f"{self.get_staging_directory(self.spec)}{os.path.basename(file)}"
            )
            self.logger.info(f"{self.spec['hostname']} Processing {current_path}")

            file_name = os.path.basename(file)

            # Handle any rename that might be specified in the spec
            if "rename" in self.spec:
                rename_regex = self.spec["rename"]["pattern"]
                rename_sub = self.spec["rename"]["sub"]

                file_name = re.sub(rename_regex, rename_sub, file_name)
                self.logger.info(
                    f"{self.spec['hostname']} Renaming file to {file_name}"
                )

            try:

                # This cannot use the standard rename, because it will fail if the file
                # is moving across filesystems. This is expected behaviour in the SFTP
                # protocol
                # Instead, we have to issue it as a command over the SSH connection
                remote_command = f"mv {current_path} {directory}/{file_name}"
                _, stdout, stderr = self.ssh_client.exec_command(remote_command)  # type: ignore[union-attr] # nosec B601
                self._log_remote_output(stdout, stderr)
                # Check the return code of the command
                remote_rc = stdout.channel.recv_exit_status()
                if remote_rc != 0:
                    self.logger.error(
                        f"[{self.spec['hostname']}] Got return code {remote_rc} from"
                        " SSH mv command"
                    )
                    return 1

                if "mode" in self.spec:
                    self.sftp_connection.chmod(
                        f"{directory}/{file_name}", int(self.spec["mode"], base=8)
                    )

                if "permissions" in self.spec:
                    # Unfortunately, this is easier to do with a proper SSH command
                    # than with the SFTP client
                    if "owner" in self.spec["permissions"]:
                        remote_command = (
                            "chown"
                            f" {self.spec['permissions']['owner']} {directory}/{file_name}"
                        )
                        _, stdout, stderr = self.ssh_client.exec_command(remote_command)  # type: ignore[union-attr] # nosec B601

                    if "group" in self.spec["permissions"]:
                        remote_command = (
                            "chgrp"
                            f" {self.spec['permissions']['group']} {directory}/{file_name}"
                        )
                        _, stdout, stderr = self.ssh_client.exec_command(remote_command)  # type: ignore[union-attr] # nosec B601

            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(
                    f"{self.spec['hostname']} Failed moving file to final location:"
                    f" {ex}"
                )
                return 1

        return 0

    def _log_remote_output(
        self, stdout: ChannelFile, stderr: ChannelStderrFile
    ) -> None:
        self.logger.info("### START OF REMOTE OUTPUT ###")
        with stdout as stdout_fh:
            str_stdout = stdout_fh.read().decode("UTF-8")
            if str_stdout:
                log_stdout(str_stdout, self.spec["hostname"], self.logger)

        with stderr as stderr_fh:
            str_stderr = stderr_fh.read().decode("UTF-8")
            if str_stderr and len(str_stderr) > 0:
                self.logger.info(
                    f"[{self.spec['hostname']}] Remote stderr returned:\n{str_stderr}"
                )

        self.logger.info("### END OF REMOTE OUTPUT ###")

    def handle_post_copy_action(self, files: list[str]) -> int:
        """Handle the post copy action specified in the config.

        Args:
            files (list[str]): A list of files that need to be handled.

        Returns:
            int: 0 if successful, 1 if not.
        """
        self.connect(self.spec["hostname"])
        sftp_client = self.ssh_client.open_sftp()  # type: ignore[union-attr]

        if self.spec["postCopyAction"]["action"] == "delete":
            # Loop through each file and use the sftp client to delete the files

            for file in files:
                try:
                    self.logger.info(f"Deleting file {file}")
                    sftp_client.remove(file)
                except OSError:
                    self.logger.error(
                        f"[{self.spec['hostname']}] Could not delete file {file} on"
                        " source host"
                    )
                    return 1

        if (
            self.spec["postCopyAction"]["action"] == "move"
            or self.spec["postCopyAction"]["action"] == "rename"
        ):
            # Use the SFTP client, and check that the destination directory exists
            move_dir = os.path.dirname(self.spec["postCopyAction"]["destination"])

            try:
                stat_result = sftp_client.stat(move_dir)
                # If it exists, then we need to ensure its a directory and not just a file
                if not stat.S_ISDIR(stat_result.st_mode):  # type: ignore[arg-type]
                    self.logger.error(
                        f"[{self.spec['hostname']}] Destination directory {move_dir} is"
                        " not a directory on source host"
                    )
                    return 1
            except OSError:
                self.logger.error(
                    f"[{self.spec['hostname']}] Destination directory {move_dir} does"
                    " not exist on source host"
                )
                return 1

                # Loop through the files and move them

            for file in files:
                try:
                    # If this is a move, then just move the file
                    if self.spec["postCopyAction"]["action"] == "move":
                        self.logger.info(
                            f"[{self.spec['hostname']}] Moving {file} to"
                            f" {self.spec['postCopyAction']['destination']}"
                        )
                        # Get the actual file name
                        file_name = os.path.basename(file)
                        sftp_client.posix_rename(
                            file,
                            f"{self.spec['postCopyAction']['destination']}/{file_name}",
                        )
                    # If this is a rename, then we need to rename the file
                    if self.spec["postCopyAction"]["action"] == "rename":
                        # Determine the new file name
                        new_file_dir = os.path.dirname(
                            self.spec["postCopyAction"]["destination"]
                        )
                        current_file_name = os.path.basename(file)

                        rename_regex = self.spec["postCopyAction"]["pattern"]
                        rename_sub = self.spec["postCopyAction"]["sub"]

                        new_file_name = re.sub(
                            rename_regex, rename_sub, current_file_name
                        )

                        self.logger.info(
                            f"[{self.spec['hostname']}] Renaming {file} to"
                            f" {new_file_dir}/{new_file_name}"
                        )
                        sftp_client.posix_rename(
                            file, f"{new_file_dir}/{new_file_name}"
                        )
                except OSError as e:
                    self.logger.error(f"[{self.spec['hostname']}] Error: {e}")
                    self.logger.error(
                        f"[{self.spec['hostname']}] Error moving or renaming file"
                        f" {file}"
                    )
                    return 1

        return 0

    def init_logwatch(self) -> int:
        """Initialise the logwatch process.

        Returns:
            int: 0 if successful, 1 if not.
        """
        self.connect(self.spec["hostname"])
        if not isinstance(self.sftp_connection, SFTPClient):
            self.logger.error(f"[{self.spec['hostname']}] Cannot connect via SFTP")
            return 1

        # There are 2 options for logwatches. One is to watch for new entries, the other is to scan the entire log.
        # Default if not specified is to watch for new entries

        # Determine the log details and check it exists first
        log_file = (
            f"{self.spec['logWatch']['directory']}/{self.spec['logWatch']['log']}"
        )

        # Stat the file
        try:
            _ = self.sftp_connection.lstat(f"{log_file}")
        except FileNotFoundError:
            self.logger.error(
                f"[{self.spec['hostname']}] Log file {log_file} does not exist"
            )
            return 1
        except PermissionError:
            self.logger.error(
                f"[{self.spec['hostname']}] Log file {log_file} cannot be accessed"
            )
            return 1

        # Open the existing file and determine the number of rows
        with self.sftp_connection.open(log_file) as log_fh:
            rows = 0
            for _, _ in enumerate(log_fh):
                pass
            self.logger.log(
                12, f"[{self.spec['hostname']}] Found {rows+1} lines in log"
            )
            self.log_watch_start_row = rows + 1

        return 0

    def do_logwatch(self) -> int:
        """Perform the logwatch process.

        Returns:
            int: 0 if successful, 1 if not.
        """
        self.connect(self.spec["hostname"])
        if not isinstance(self.sftp_connection, SFTPClient):
            self.logger.error(f"[{self.spec['hostname']}] Cannot connect via SFTP")
            return 1

        # Determine if the config requires scanning the entire log, or just from the start_row determine in the init function
        start_row = (
            self.log_watch_start_row
            if "tail" in self.spec["logWatch"] and self.spec["logWatch"]["tail"]
            else 0
        )
        self.logger.log(
            12, f"[{self.spec['hostname']}] Starting logwatch from row {start_row}"
        )

        # Open the remote log file and parse each line for the pattern
        log_file = (
            f"{self.spec['logWatch']['directory']}/{self.spec['logWatch']['log']}"
        )

        with self.sftp_connection.open(log_file) as log_fh:
            for i, line in enumerate(log_fh):
                # We need to start after the previous line in the log
                if i >= start_row:
                    self.logger.log(
                        11, f"[{self.spec['hostname']}] Log line: {line.strip()}"
                    )
                    if re.search(self.spec["logWatch"]["contentRegex"], line.strip()):
                        self.logger.log(
                            12,
                            f"[{self.spec['hostname']}] Found matching line in log:"
                            f" {line.strip()} on line: {i+1}",
                        )
                        return 0

        return 1

    def create_flag_files(self) -> int:
        """Create the flag files on the remote host.

        Returns:
            int: 0 if successful, 1 if not.
        """
        self.connect(self.spec["hostname"])

        if not isinstance(self.ssh_client, SSHClient):
            self.logger.error(f"[{self.spec['hostname']}] Cannot connect via SSH")
            return 1

        # Manually open SFTP client
        sftp_client = self.ssh_client.open_sftp()
        filename = self.spec["flags"]["fullPath"]

        try:
            # Use the SFTP client to create an empty file at this path
            sftp_client.file(filename, "w").close()

            # Set permissions on the file to whatever was specified in the spec,
            # otherwise we leave them as is
            # We cannot change ownership without using sudo, so we don't bother
            if "permissions" in self.spec:
                sftp_client.chmod(filename, int(self.spec["permissions"], base=8))

        except OSError as e:
            self.logger.error(f"[{self.spec['hostname']}] Error: {e}")
            self.logger.error(
                f"[{self.spec['hostname']}] Error creating flag file: {filename}"
            )
            return 1

        return 0


def log_stdout(str_stdout: str, hostname: str, logger: logging.Logger) -> None:
    """Log the stdout from a remote command in a nice format.

    Args:
        str_stdout (str): The stdout from the remote command
        hostname (str): The hostname of the remote host
        logger (logging.Logger): The logger to use
    """
    for line in str_stdout.splitlines():
        logger.info(f"[{hostname}] REMOTE OUTPUT: {line}")


def mkdir_p(sftp: SFTPClient, remote_directory: str) -> bool:
    """Change to this directory, recursively making new folders if needed.

    Thanks to: https://stackoverflow.com/a/14819803

    Args:
        sftp (SFTPClient): The SFTP client to use
        remote_directory (str): The remote directory to create

    Returns:
        bool: True if successful, False if not.
    """
    if remote_directory == "/":
        # absolute path so change directory to root
        sftp.chdir("/")
        return False
    if remote_directory == "":
        # top-level relative directory must exist
        return False
    try:
        sftp.chdir(remote_directory)  # sub-directory exists
    except FileNotFoundError:
        dirname, basename = os.path.split(remote_directory.rstrip("/"))
        mkdir_p(sftp, dirname)  # make parent directories
        sftp.mkdir(basename)  # sub-directory missing, so created it
        sftp.chdir(basename)
        return True

    return False


class SSHExecution(RemoteExecutionHandler):
    """Class to handle SSH execution of remote commands."""

    TASK_TYPE = "E"
    ps_regex = r"(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(.*)"
    ssh_client: SSHClient
    remote_pid: int

    def tidy(self) -> None:
        """Tidy up the SSH connection."""
        if self.ssh_client:
            self.logger.info(f"[{self.remote_host}] Closing SFTP connection")
            self.ssh_client.close()

    def __init__(self, remote_host: str, spec: dict):
        """Initialise the SSHExecution class.

        Args:
            remote_host (str): The hostname of the remote host
            spec (dict): The specification for the remote command
        """
        self.remote_host: str = remote_host
        self.random: int = random.randint(
            100000, 999999
        )  # Random number used to make sure when we kill stuff, we always kill the right thing

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        super().__init__(spec)

        client = SSHClient()
        client.set_log_channel(f"{__name__}.{ spec['task_id']}.paramiko.transport")
        client.set_missing_host_key_policy(AutoAddPolicy())

        self.ssh_client = client

    def connect(self) -> None:
        """Connect to the remote host."""
        if self.ssh_client and isinstance(self.ssh_client, SSHClient):
            transport = self.ssh_client.get_transport()
            if (
                transport is not None
                and isinstance(transport, Transport)
                and transport.is_active()
            ):
                return

        kwargs = {
            "hostname": self.remote_host,
            "port": (self.spec["protocol"].get("port", 22)),
            "username": self.spec["protocol"]["credentials"]["username"],
            "timeout": 5,
        }

        # If a custom key is set via env vars, then set that
        if (
            os.environ.get("OTF_SSH_KEY")
            and os.path.exists(str(os.environ.get("OTF_SSH_KEY")))
        ) and "keyFile" not in self.spec["protocol"]["credentials"]:
            self.logger.info("Loading custom private SSH key from OTF_SSH_KEY env var")
            key = RSAKey.from_private_key_file(str(os.environ.get("OTF_SSH_KEY")))
            kwargs["pkey"] = key

        # If a specific key file has been defined, then use that
        elif "keyFile" in self.spec["protocol"]["credentials"]:
            self.logger.info("Using key file from task spec")
            kwargs["key_filename"] = self.spec["protocol"]["credentials"]["keyFile"]

        try:
            self.ssh_client.connect(**kwargs)
            _, stdout, _ = self.ssh_client.exec_command("uname -a")  # nosec B601
            with stdout as stdout_fh:
                output = stdout_fh.read().decode("UTF-8")
                self.logger.log(11, f"[{self.remote_host}] Remote uname: {output}")
        except Exception as ex:
            self.logger.error(f"Unable to connect to {self.remote_host}: {ex}")
            raise ex

    def _get_child_processes(self, parent_pid: int, process_listing: list) -> list:
        """Get the child processes of a given PID.

        Args:
            parent_pid (int): The PID of the parent process
            process_listing (list): The list of processes running on the remote host

        Returns:
            list: A list of child PIDs
        """
        children = []
        for line in process_listing:
            match = re.search(self.ps_regex, line)
            if match:
                if int(match.group(3)) == parent_pid:
                    child_pid = int(match.group(2))
                    # Never add PID 1 or 0!
                    if child_pid in (1, 0):
                        continue
                    self.logger.debug(
                        f"[{self.remote_host}] Found child process with PID:"
                        f" {child_pid}"
                    )
                    children.append(child_pid)
                    # Recurse to find the children of this child
                    children.extend(
                        self._get_child_processes(child_pid, process_listing)
                    )
        return children

    def kill(self) -> None:
        """Kill the remote process."""
        self.logger.info(f"[{self.remote_host}] Killing remote process")

        self.connect()
        # We know the top level remote PID, we need to get all the child processes associated with it
        _, stdout, _ = self.ssh_client.exec_command("ps -ef")  # nosec B601
        process_listing = []
        # Get the process listing
        with stdout as stdout_fh:
            process_listing = stdout_fh.read().decode("UTF-8").splitlines()

        # Now we have this, parse it, find the parent PID, and then all the children
        children = self._get_child_processes(self.remote_pid, process_listing)
        children.append(self.remote_pid)
        self.logger.info(
            f"[{self.remote_host}] Found {len(children)} child processes to kill -"
            f" {children}"
        )

        # Now we have the list of children, kill them
        command = f"kill {' '.join([str(x) for x in children])}"
        self.logger.info(
            f"[{self.remote_host}] Killing remote processes with command: {command}"
        )

        # Make sure we're not killing PID 1 or 0
        if command in ("kill 1", "kill 0"):
            self.logger.error(
                f"[{self.remote_host}] Refusing to kill PID 1 or 0, aborting"
            )
            return
        _, stdout, _ = self.ssh_client.exec_command(command)  # nosec B601
        # Wait for the command to finish
        while not stdout.channel.exit_status_ready():
            time.sleep(0.1)

        # Disconnect SSH
        self.tidy()

    def execute(self) -> bool:
        """Execute the remote command.

        Returns:
            bool: True if the command was executed successfully, False otherwise
        """
        try:
            self.connect()

            directory = quote(self.spec["directory"])

            command = (
                f"echo __OTF_TOKEN__$$_{self.random}__; cd {directory} &&"
                f" {self.spec['command']}"
            )

            self.logger.info(f"[{self.remote_host}] Executing command: {command}")

            _, stdout, stderr = self.ssh_client.exec_command(command)  # nosec B601

            # Log the stdout and stderr
            for line in iter(lambda: str(stdout.readline(2048)), ""):
                log_stdout(line, self.remote_host, self.logger)

                # Check the line for the token and pull out the PID
                regex = f"__OTF_TOKEN__(\\d+)_{self.random}__"
                pid_search = re.search(regex, line)
                if pid_search:
                    self.remote_pid = int(pid_search.group(1))
                    self.logger.info(
                        f"[{self.remote_host}] Found remote PID: {self.remote_pid}"
                    )

            with stderr as stderr_fh:
                str_stderr = stderr_fh.read().decode("UTF-8")
                if str_stderr and len(str_stderr) > 0:
                    self.logger.info(
                        f"[{self.remote_host}] Remote stderr returned:\n{str_stderr}"
                    )

            # Get the return code
            remote_rc = stdout.channel.recv_exit_status()
            self.tidy()
            if remote_rc != 0:
                self.logger.error(
                    f"[{self.remote_host}] Got return code {remote_rc} from SSH command"
                )
                return False

            return True
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.logger.error(f"[{self.remote_host}] Exception caught: {e}")
            return False



"""Email handler to send files via email."""

import glob
import smtplib
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from logging import DEBUG

import opentaskpy.otflogging
from opentaskpy.remotehandlers.remotehandler import RemoteTransferHandler

MAX_OBJECTS_PER_QUERY = 100


class EmailTransfer(RemoteTransferHandler):
    """A remote handler for email transfers."""

    TASK_TYPE = "T"

    def __init__(self, spec: dict):
        """Initialise the handler."""
        self.protocol_vars: dict

        self.logger = opentaskpy.otflogging.init_logging(
            __name__, spec["task_id"], self.TASK_TYPE
        )

        # Ensure this logs to the same location as this module
        opentaskpy.otflogging.set_log_file("smtplib")

        super().__init__(spec)

    def supports_direct_transfer(self) -> bool:
        """Return whether this handler supports direct transfers."""
        return False

    def set_handler_vars(self, protocol_vars: dict) -> None:
        """Set the handler variables.

        Set any custom variables that are specifically needed for this handler.

        The schema defines the following variables:
        - smtp_server - The SMTP server to use
        - smtp_port - The SMTP port to use
        - sender - The sender of the email
        - credentials - A dictionary containing the following:
            - username - The username used to authenticate
            - password - The password used to authenticate

        Args:
            protocol_vars (dict): The protocol variables.

        """
        self.protocol_vars = protocol_vars

        # Also pull variables that have been set on the spec level
        # Replace/Append anything defined in self.protocol_vars with anything in self.spec["protocol"]
        for key, value in self.spec["protocol"].items():
            self.protocol_vars[key] = value
        # Remove name
        del self.protocol_vars["name"]

    def push_files_from_worker(
        self, local_staging_directory: str, file_list: dict | None = None
    ) -> int:
        """Push files from the worker to the email recipients.

        Args:
            local_staging_directory (str): The local staging directory.
            file_list (dict, optional): A dictionary of files to transfer. Defaults to None.

        Returns:
            int: The result of the transfer.
        """
        result = 0
        if file_list:
            files = list(file_list.keys())
        else:
            files = glob.glob(f"{local_staging_directory}/*")

        for email_address in self.spec["recipients"]:
            # Create an email message
            msg = MIMEMultipart()

            # Attach the files to the message
            for file in files:
                # Strip the directory from the file
                file_name = file.split("/")[-1]
                self.logger.debug(f"Emailing file: {files} to {email_address}")
                try:
                    with open(file, "rb") as file_handle:
                        part = MIMEApplication(file_handle.read(), Name=file_name)
                    # After the file is closed
                    part["Content-Disposition"] = f'attachment; filename="{file_name}"'
                    msg.attach(part)
                except Exception as ex:  # pylint: disable=broad-exception-caught
                    self.logger.error(f"Failed to attach file: {file}")
                    self.logger.error(ex)
                    result = 1

            # Get comma separated list of files
            attachment_file_list = ", ".join([file.split("/")[-1] for file in files])

            # Add a plaintext body to the email
            msg.attach(
                MIMEText(
                    self.spec.get(
                        "message", f"Please find attached: {attachment_file_list}"
                    )
                )
            )
            # Set the email subject
            if "subject" in self.spec:
                msg["Subject"] = self.spec["subject"]

            msg["From"] = self.protocol_vars["sender"]

            # Send the email using a provided SMTP server
            try:
                self.logger.debug(f"Sending email to {email_address}")
                smtp = smtplib.SMTP(
                    self.protocol_vars["smtp_server"],
                    port=self.protocol_vars["smtp_port"],
                )
                if self.logger.getEffectiveLevel() <= DEBUG:
                    smtp.set_debuglevel(1)
                smtp.starttls()

                # Authenticate
                smtp.login(
                    self.protocol_vars["credentials"]["username"],
                    self.protocol_vars["credentials"]["password"],
                )

                smtp.sendmail(
                    self.protocol_vars["sender"], email_address, msg.as_string()
                )
                smtp.quit()
            except Exception as ex:  # pylint: disable=broad-exception-caught
                self.logger.error(f"Failed to send email to {email_address}")
                self.logger.error(ex)
                result = 1

        return result

    def pull_files_to_worker(self, local_staging_directory: str) -> int:
        """Not implemented for this handler."""
        raise NotImplementedError

    def handle_post_copy_action(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def list_files(
        self, directory: str | None = None, file_pattern: str | None = None
    ) -> dict:
        """Not implemented for this handler."""
        raise NotImplementedError

    def move_files_to_final_location(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def pull_files(self, files: list) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError

    def transfer_files(
        self, files: dict, remote_spec: dict, dest_remote_handler: dict | None = None
    ) -> None:
        """Not implemented for this handler."""
        raise NotImplementedError



"""An example plugin that simply returns a random number between the 2 provided values."""

import random

import opentaskpy.otflogging
from opentaskpy.exceptions import LookupPluginError

logger = opentaskpy.otflogging.init_logging(__name__)

PLUGIN_NAME = "random_number"


def run(**kwargs) -> str:  # type: ignore[no-untyped-def]
    """Returns a random number between the 2 provided values.

    Args:
        **kwargs: Expect a kwarg named min and max

    Raises:
        LookupPluginError: Returned if the kwarg 'min' or 'max' is not provided
        FileNotFoundError: Returned if the file does not exist

    Returns:
        _type_: The value read from the file
    """
    # Expect a kwarg named min, and max
    expected_kwargs = ["min", "max"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise LookupPluginError(
                f"Missing kwarg: '{kwarg}' while trying to run lookup plugin"
                f" '{PLUGIN_NAME}'"
            )

    # Return a random number between the 2 provided values
    return str(random.randint(kwargs["min"], kwargs["max"]))
